// basisfilter.C 8/97  Steve Ludtke

//N basisfilter
//F This takes an ortogonal basis set as created by svdcmp and uses it to operate on a set of images
//T LM
//1
//P <input>	Input images to process
//P <basis set>	A set of orthogonal basis images as generated by svdcmp
//P [frac=<num>/<denom>]	Operate only on a fraction of the raw images
//P [nbasis=<n>]	Number of basis images to use from the basis set. Defaults to all.
//P [nbasiscls=<n>]	Number of basis images to use for classification. Defaults to nbasis.
//P [clsbylcmp]	Classify using 'lcmp', linear comparison in real space (without basis projection)
//P [filter=<output>]	Uses the basis set as a filter on the data
//P [classify=<ref set>]	Performs classification of the input images, <ref set> should be projections to use for classification.
//P [align=<ref set>]	Align input images to the best match from <ref set>, should be same set used to generate <basis set>
//P [multiali=<n>]	Rather than just taking the best 'align' match, this will take the several best matches and select the one with the strongest decomposition
//P [poweralign[=<max dx>]]	Align particles by projected power maximization
//P [saveali=<output>]	Saves aligned unfiltered particles in <output>
//P [origrt]	If the inputs are aligned before filtration, this will cause the filtered output to be returned to the original orienatation
//P [scalebasis]	Scale the basis vectors by their significance in <ref set>
//P [verbose]	More verbose output
//U basisfilter start.hed msa.hed start.filt.hed align=proj.hed origrt
//D This program performs various operations on a set of (noisy) input images using an orthogonal
//E basis set.  The basis vectors are generated by singlar value decomposition/multivariate 
//E statistical analysis/prinicpal component analysis (effectively different
//E terms for the same thing in this context). Generally this set is produced using the program 'svdcmp'.
//D When used to filter data, this program will produce images 
//E with dramatically less noise, however, somewhat like a low-pass filter,
//E substantial real information will be missing from the filtered images, particularly at high resolution. The input
//E images are generally a set of projections of a 3D model. If this 3D model is substantially different from the
//E model the noisy images derive from, this filtration could induce significant artifacts.
//D This routine is mainly aimed at providing less noisy particles for use in alignment/orientation determination.
//E Particles filtered through this program should not be used as actual data for reconstruction.


#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "XYData.h"

main(int argc, char *argv[]) {
int i,j,num=0,denom=1,clsmeth=0;
int alignn=1,origrt=0,nbasis=-1,nbasisfilt=-1,nbasiscls=-1,scalebasis=0,verbose=0,powali=0;
float dang=PI;
char *alignfile=NULL,*clsref=NULL,*filterout=NULL,*saveali=NULL;


appinit(argc,argv);

if (argc<3) appusage(argv[0]);


for (i=3; i<argc; i++) {	
	if (strncmp(argv[i],"align=",6)==0) alignfile=&argv[i][6];
	else if (strncmp(argv[i],"multiali=",9)==0) alignn=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"nbasis=",7)==0) nbasisfilt=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"nbasiscls=",10)==0) nbasiscls=atoi(&argv[i][10]);
	else if (strncmp(argv[i],"filter=",7)==0) filterout=&argv[i][7];
	else if (strncmp(argv[i],"origrt",6)==0) origrt=1;
	else if (strncmp(argv[i],"classify=",9)==0) { clsref=&argv[i][9]; if (!clsmeth) clsmeth=2; }
	else if (strncmp(argv[i],"clsbylcmp",9)==0) clsmeth=1;
	else if (strncmp(argv[i],"saveali=",8)==0) saveali=&argv[i][8];
	else if (strncmp(argv[i],"usefs=",6)==0) ;
	else if (strncmp(argv[i],"scalebasis",10)==0) scalebasis=1;
	else if (strncmp(argv[i],"poweralign",10)==0) {
		powali=-1;
		if (argv[i][10]=='=') powali=atoi(&argv[i][11]);
	}
	else if (strncmp(argv[i],"verbose",7)==0) {
		if (argv[i][7]!='=') verbose=1;
		else verbose=atoi(&argv[i][8]);
	}
	else { printf("Unknown option %s\n",argv[i]); exit(1); }
}

if (nbasisfilt>nbasiscls) nbasis=nbasisfilt; else nbasis=nbasiscls;
if (nbasiscls==-1) nbasiscls=nbasis;
if (nbasisfilt==-1) nbasisfilt=nbasis;

if (!alignfile && alignn>1) error(ERR_EXIT,"multiali specified without align");

int bc=fileCount(argv[2]);
if (bc<2) error(ERR_EXIT,"Need at least 2 basis images!");
if (nbasis<0) nbasis=bc-1;
if (nbasis>bc) { error(ERR_WARN,"nbasis > # images in basis file"); nbasis=bc-1; }

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);
LOG(Ref,argv[2],LOG_REFFILE,NULL);

// read basis set and normalize
List *basis=readImages(argv[2],0,nbasis-1);
for (i=0; i<nbasis; i++) {
	EMData *tmp=((EMData *)basis->objectAt(i));
	tmp->edgeNormalize();
	tmp->multConst(1.0/tmp->SumSq());	// a 'true' normalization
}
if (powali<0) powali=((EMData *)basis->objectAt(0))->ySize()/8;
dang=atan2(1.0,((EMData *)basis->objectAt(0))->ySize()/3.0);

// read classification refs if present, and determine basis projections
int nclsref=-1;
if (clsref) {
	nclsref=fileCount(clsref); 
	if (nclsref<2) error(ERR_EXIT,"Could not read 2 or more references from classify file");
}
else nclsref=1;

//float clsvec[nbasis][nclsref];		// GCC only extension
float **clsvec;
clsvec=(float **)malloc(nbasis*sizeof(float*));
for (i=0; i<nbasis; i++) clsvec[i]=(float *)malloc(nclsref*sizeof(float)); 

float *clssig = (float *)malloc(nbasis*sizeof(float));
List clsrefs;
// If classification is going to occur, we need to do some initialization
if (clsref) {
	if (verbose) printf("Read classification refs (%s)\n",clsref);

	float *clsmean = (float *)malloc(nbasis*sizeof(float));
	for (j=0; j<nbasis; j++) clsmean[j]=clssig[j]=0;

	// outer loop over classification refs
	for (i=0; i<nclsref; i++) {
		//set up cls files
/*		char s[200];
		sprintf(s,"cls%04d.lst",i);
		unlink(s);
		FILE *out=fopen(s,"w");
		fprintf(out,"#LST\n");
		fclose(out);*/

		// read reference
		EMData in;
		in.readImage(clsref,i);
		float norm=0;
		// determine basis projection for each ref
		for (j=0; j<nbasis; j++) {
			clsvec[j][i]=in.dot((EMData *)basis->objectAt(j));
			norm+=SQR(clsvec[j][i]);
		}
		// normalize basis vector
		for (j=0; j<nbasis; j++) clsvec[j][i]/=sqrt(norm);

		// We get the std. dev. of each basis component
		for (j=0; j<nbasis; j++) {
			clsmean[j]+=clsvec[j][i];
			clssig[j]+=SQR(clsvec[j][i]);
		}

		if (clsmeth==1) clsrefs.addObject(in.copy(0,0));

		if (verbose>1) {
			printf("%d.\t",i);
			for (j=0; j<nbasis; j++) printf("%1.4f\t",clsvec[j][i]);
			printf("\n");
		}
	}
	
	// Print out the sigma values for each basis component
	if (verbose>1) printf("S.\t");
	for (j=0; j<nbasis; j++) {
		clsmean[j]/=nclsref;
		clssig[j]=sqrt(clssig[j]/nclsref-SQR(clsmean[j]));
		if (verbose>1) printf("%1.4f\t",clssig[j]);
	}
	if (verbose>1) printf("\n\n");
	
/*
	if (scalebasis) {
		// scale the basis vectors
		for (i=0; i<nbasis; i++) ((EMData *)basis->objectAt(i))->multConst(clssig[i]);

		// now we redo this. Silly, but...
		for (i=0; i<nclsref; i++) {
			EMData in;
			in.readImage(clsref,i);
			float norm=0;
			// determine basis projection for each ref
			for (j=0; j<nbasis; j++) {
				clsvec[j][i]=in.dot((EMData *)basis->objectAt(j));
				norm+=SQR(clsvec[j][i]);
			}
			// normalize basis vector
			for (j=0; j<nbasis; j++) clsvec[j][i]/=sqrt(norm);

			// print out the new scaled reference vectors
			if (verbose>1) {
				printf("%d.\t",i);
				for (j=0; j<nbasis; j++) printf("%1.4f\t",clsvec[j][i]);
				printf("\n");
			}
		}
	}*/
	free(clsmean);
}

if (verbose) printf("Begin analysis:\n");
// Image filtration
// outer loop over images being filtered
int nimg=fileCount(argv[1]);
int start=nimg*num/denom;
int stop=nimg*(num+1)/denom;
for (i=start; i<stop; i++) {
	float dts=0;

	float *vec = (float*)malloc(nbasis*sizeof(float));

	EMData in;
	in.readImage(argv[1],i);
	in.edgeNormalize();

	// rot/trans alignment my maximizing basis projection power
	if (powali) {
		EMData *cop=in.copy();
		float x,y,bx,by,bx2,by2;
		int nx=in.xSize(),ny=in.ySize();
		int fl,bfl=0;
		float ang=0,bang,bang2,bpow=0;
		for (fl=0; fl<2; fl++) {
			for (ang=0; ang<PI*2; ang+=dang*4.0) {	//4
				cop->setRAlign(ang);
				cop->setTAlign(0,0);
				cop->rotateAndTranslate();
				if (fl) cop->vFlip();
				EMData *ccf,*ccfs;
				ccfs=cop->copyHead();
				ccfs->zero();
				for (j=0; j<nbasisfilt; j++) {
					ccf=cop->calcCCF((EMData *)basis->objectAt(j));
					ccf->realFilter(18);
					ccfs->add(ccf);
					delete ccf;
				}
				ccfs->toCorner();
				for (y=ny/4; y<ny*3/4; y++) {
					for (x=nx/4; x<nx*3/4; x++) {
						if (ccfs->valueAt(x,y)>bpow) { bpow=ccfs->valueAt(x,y); bang=ang; bx=nx/2-x; by=ny/2-y;  bfl=fl; }
					}
				}
				delete ccfs;
			}
		}
		// corrects for inverting the order of r/t above		
		if (bfl) by*=-1.0;
		x= bx*cos(bang)+by*sin(bang);
		y=-bx*sin(bang)+by*cos(bang);
		bx=x;
		by=y;

		bpow=0;		// the normalization won't match, so start from scratch
		bang2=bang;
		// now refine the dx/dy and rotation
		x=bx;
		y=by;
		for (ang=bang2-dang*4.0; ang<bang2+dang*4.0; ang+=dang/2.) {
			cop->setRAlign(ang);
			cop->setTAlign(x,y,0);
			cop->rotateAndTranslate();
			if (bfl) cop->vFlip();
			float pow=0;
			for (j=0; j<nbasisfilt; j++) pow+=SQR(cop->dot((EMData *)basis->objectAt(j)));
			if (pow>bpow) { bpow=pow; bang=ang; bx=x; by=y; }
		}

		bpow=0;		// the normalization won't match, so start from scratch
		bx2=bx;
		by2=by;
		bang2=bang;
		// now refine the dx/dy and rotation
		for (x=bx2-.75; x<=bx2+.76; x+=.25) {
			for (y=by2-.75; y<=by2+.76; y+=.25) {
				for (ang=bang2-dang; ang<bang2+dang; ang+=dang/4.) {
					cop->setRAlign(ang);
					cop->setTAlign(x,y,0);
					cop->rotateAndTranslate();
					if (bfl) cop->vFlip();
					float pow=0;
					for (j=0; j<nbasisfilt; j++) pow+=SQR(cop->dot((EMData *)basis->objectAt(j)));
					if (pow>bpow) { bpow=pow; bang=ang; bx=x; by=y; }
				}
			}
		}
		delete cop;
		if (verbose>1) printf("%d.\t%0.1f\t%0.1f\t%1.1f\t%c\n",i,bx,by,bang*180./PI,bfl?'F':' ');
		in.setRAlign(bang);
		in.setTAlign(bx,by,0);
		in.rotateAndTranslate();
		if (bfl) in.vFlip();
	}

	if (saveali) in.writeIMAGIC(saveali,i);

	// inner loop dot product over basis vectors
	for (j=0; j<nbasis; j++) vec[j]=in.dot((EMData *)basis->objectAt(j));

	// This does the classification
	if (clsref) {
		float max=-1.0e30;
		int maxj=0;
		// classify based on lcmp between ref and particle
		if (clsmeth==1) {
			float cmp;
			for (j=0; j<nclsref; j++) {
				cmp=10.0-((EMData *)(clsrefs.objectAt(j)))->lcmp(&in,1);
				if (cmp>max) { max=cmp; maxj=j; }
			}
		}
		// classify based on basis projection comparisons
		float *ddt=(float *)malloc(nclsref*sizeof(float));
		if (clsmeth==2) {
		
			// similarity dot product for each possibility
			for (j=0; j<nclsref; j++) {
				ddt[j]=0;
				if (scalebasis) {
					for (int k=0; k<nbasiscls; k++) ddt[j]+=vec[k]*clsvec[k][j]*clssig[k];
				} else {
					for (int k=0; k<nbasiscls; k++) ddt[j]+=vec[k]*clsvec[k][j];
				}
				if (ddt[j]>max) { max=ddt[j]; maxj=j; }
			}
		}

		char s[200],t[200];
		sprintf(s,"cls%04d.lst",maxj);
		sprintf(t,"%f",max);
		if (saveali) in.writeLST(s,-1,saveali,i,t);
		else in.writeLST(s,-1,argv[1],i,t);
		
		if (verbose==1) printf("%d -> %d  (%1.4g)\n",i,maxj,max);
		if (verbose>1) {
			printf("%d.",i);
			for (j=0; j<nclsref; j++) printf("\t%1.4f",ddt[j]);
			printf(" -> %d\n",maxj);
		}
		free(ddt);
	}

	// This does the basis filter
	if (filterout) {
		EMData *out=in.copyHead();
		out->zero();
		for (j=0; j<nbasisfilt; j++){
			EMData *tmp=((EMData *)basis->objectAt(j))->copy(0,0);
			
			float dt=vec[j];
			tmp->multConst(dt);
			out->add(tmp);
			delete tmp;
			dts+=dt*dt;
		}
		out->writeIMAGIC(filterout,i);
		delete out;

		if (verbose) {
			printf("%d.\t",i);	
			if (verbose>1) { for (j=0; j<nbasisfilt; j++) printf("%1.3f\t",vec[j]); }
			if (filterout) printf("%1.4g\n",dts);
		}
	}
	
	free(vec);
	
}
LOG(Ref,0,LOG_END,NULL);

}
