// make3diter2.C 2/01 Steve Ludtke
// This program takes a set of class averages and a 3D Fourier model and
// iteratively refines it in real space, that is, it tries to make
// the model match the projections better

//N make3diter2
//F Iteratively improves 3D models using real-space methods.
//T LR
//1
//P <3d map in>	The input map to be improved
//P <class averages>	The class averages used to construct the model
//P <3d map deriv>	The derivput, same as input ok
//P [sym=<c4,icos,etc>]	Symmetry to use
//P [minpart=<n ptcl>]	Specify minimum number of particles in a class-average.
//U make3diter2 threed.3.mrc classes.3.img threed.3a.mrc sym=c4
//D This program will take a set of class averages, and a 3D model generated by
//E make3d, and attempt to eliminate Fourier artifacts using an iterative real-space
//E technique. <a href=make3diter.html>make3diter<a> seems to work better.

#include <stdio.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "Euler.h"
#include <math.h>
#include <ctype.h>

struct FIX { float diff; int n; };

int cmp(const void *f1, const void *f2) { 
	float f = ((FIX *)f1)->diff - ((FIX *)f2)->diff;
	if (f<0) return -1;
	if (f>0) return 1;
	return 0;
}

main(int argc, char *argv[]) {
char s[MAXPATHLEN];
int nsym=1;
int i,j,k,l,n,nx,np,npt,pn0,pn1,minpart=0;
float x,y,z,sig=0,scale=1.0;
float x2,y2,r;
List *proj;
float **mx;
FIX *pred;
EMData *d1,*c,*p;
float *dt1;
char *sym=NULL;

appinit(argc,argv);

if (argc<4) appusage(argv[0]);

for (i=4; i<argc; i++) {	
	if (strncmp(argv[i],"sym=",4)==0) {
                sym=&argv[i][4];
		Euler e(0,0,0);
		e.setSym(sym);
		nsym=e.getMaxSymEl();
                if (e.getMaxSymEl()==0) { printf("Unsupported symmetry\n"); exit(1); }
	}
	else if (strncmp(argv[i],"minpart=",8)==0) minpart=atoi(&argv[i][8]);
	else if (strncmp(argv[i],"scale=",6)==0) scale=atof(&argv[i][6]);
	else { printf("Unknown argument '%s'\n",argv[i]); exit(1); }
}

int Ref;
Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);
LOG(Ref,argv[2],LOG_INFILE,NULL);


// read class averages
proj=readImages(argv[2],0,-1);
n=proj->count();
printf("%d images read\n",n);

// strip deriv references
npt=0;
for (i=n-1; i>=0; i--) {
	c=(EMData *)proj->objectAt(i);
	if (c->NImg()<minpart) { proj->removeObjectAt(i); delete c; }
	npt+=c->NImg();
}
npt*=nsym;
n=proj->count();
printf("Using %d class averages\n",n);

if (n<4) { printf("Input file too small ?!?!\n"); exit(1); }
mx=(float**)malloc(n*sizeof(float *)*nsym);
pred=(FIX *)malloc(n*sizeof(FIX)*nsym);


// read 3d model
d1=new EMData;
d1->readImage(argv[1],-1);
if (d1->xSize() != d1->ySize() || d1->ySize() != d1->zSize()) {
	printf("Sorry, nx, ny, nz must be equal\n");
	exit(1);
}
d1->edgeNormalize(1);
d1->applyMask(-1,4);
nx=d1->xSize();


// now subtract real space projections from each slice
printf("Determining projection differences\n");
for (i=0; i<n; i++) {
	c=(EMData *)proj->objectAt(i);
	p=d1->project3d(c->alt(),c->az(),c->phi(),-1);

	// renormalize class-average
	c->edgeNormalize(1);
	c->applyMask(-1,4);
//	c->edgeNormalize(1);
	c->normalizeTo(p);
	c->subtract(p);
	delete p;
c->writeIMAGIC("tmp.hed",-1);
	printf("%d.\t%1.3f         \r",i,c->Sigma()); fflush(stdout);

	sig+=c->Sigma()*c->NImg();
}

printf("\nMean sigma of differences (%f) = %1.4f\n\n",scale,sig/(npt*d1->Sigma()));

d1->setRAlign(0,0,0);
d1->setTAlign(0,0,0);
d1->rotateAndTranslate();

//Build a 3D derivative model
EMData *deriv=new EMData,*d;
deriv->setSize(d1->xSize(),d1->xSize(),d1->xSize());		// use whatever the last image is, all should be the same
deriv->zero();

float norm=0;
for (i=0; i<proj->count(); i++) {
	d=(EMData *)proj->objectAt(i);
	if (d->NImg()<0) continue;
	deriv->insertSlice3DReal(d,d->NImg());
	norm+=d->NImg();
	printf("%d. %f\t%f\t%f\t%d\n",i,d->alt()*180.0/PI,d->az()*180.0/PI,d->phi()*180.0/PI,d->NImg());
}
deriv->multConst(1.0/(norm*d->xSize()));

// Apply symmetry if it was specified
if (sym) {
	Euler e,*a;
	EMData *d,*d2;
	int ns=0;

	d=deriv->copy();
	d->setParent(NULL);
	d2=d->copy();

	printf("Applying %s symmetry\n",sym);
	a=d1->getEuler();
	a->setAngle(0,0,0);
	a->setSym(sym);
	deriv->zero();
	for (e=a->FirstSym(); e.valid(); e=a->NextSym(),ns++) {
		printf("Angle %f,%f,%f\n",e.alt()*180/PI,e.az()*180/PI,e.phi()*180/PI);
		d2->setRAlign(e.alt(),e.az(),e.phi());
		d2->rotateAndTranslate();
		deriv->add(d2);
	}
	deriv->multConst(1.0/ns);
	delete d;
	delete d2;
}
deriv->multConst(-1.0*scale);
deriv->writeMRC("deriv.mrc");

d1->add(deriv);
d1->writeMRC(argv[3]);

LOG(Ref,argv[3],LOG_OUTFILE,NULL);
LOG(Ref,NULL,LOG_END,s);
}
