// svdcmp  1/20/02  Steve Ludtke
// Copyright 2002  Baylor College of Medicine

//N svdcmp
//F This will determine a set of orthogonal basis images from a set of aligned particles
//T LA
//1
//P <images>	Input file, images to be analyzed
//P <basis>	Output file, set of orthogonal basis images
//P [n=<n basis>]	Number of basis images to output (<= number of input images)
//P [mask=<radius>]	Mask radius in pixels (use the smallest reasonable value)
//P [jacobi]	Uses Jacobi method, much slower, but more accurate
//U svdcmp proj.hed basis.hed
//D This program is designed to produce a set of orthogonal basis images to
//E use for noise reduction on raw particles. Typically this program will
//E be used on a set of projections covering the asymmetric triangle, but
//E with no in-plane rotation. The resulting basis images can then be applied
//E to an (aligned) noisy image to act as a sort of filter. 

#include <stdio.h>
#include <math.h>
#include <gsl/gsl_linalg.h>
#include "util.h"
#include "EMData.h"

#define GSL_RANGE_CHECK_OFF

//#define N 200
//#define M 20000

int circularcount(int r);
void circularpack(EMData *src,gsl_matrix *dest,int mxy,int r);
void circularunpack(gsl_matrix *src, EMData *dest, int mxy,int r);

main(int argc,char *argv[]) {
int i,j;
int mask=-1,nbasis=-1,imsize=-1,jacobi=0;
EMData data;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);
for (i=3; i<argc; i++) {        
	if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"n=",2)==0) nbasis=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"jacobi",6)==0) jacobi=1;
	else if (strncmp(argv[i],"usefs",5)==0) ;
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

int nimg=fileCount(argv[1]);
if (nimg==0) error(ERR_EXIT,"No readable images in input file");

{
	EMData in;
	in.readImage(argv[1],0,1);
	imsize=in.ySize();
	if (mask<=0) mask=in.ySize()/2-1;
}

if (nbasis<=0) nbasis=nimg;
int pixels=circularcount(mask);

// If the matrix is over 100 megs, warn the user
if (pixels*nimg*sizeof(float)>100000000) error(ERR_WARN,"This is a VERY large matrix. It will take quite some time.");

// Allocate the working space
gsl_vector *work=gsl_vector_alloc(nimg);
gsl_vector *S=gsl_vector_alloc(nimg);
gsl_matrix *A=gsl_matrix_alloc(pixels,nimg);
gsl_matrix *V=gsl_matrix_alloc(nimg,nimg);
gsl_matrix *X=gsl_matrix_alloc(nimg,nimg);

// Pack the image data into the input matrix
for (i=0; i<nimg; i++) {
	EMData in;
	in.readImage(argv[1],i);
	circularpack(&in,A,i,mask);
}

// Do the decomposition. All the real work is here
if (!jacobi) gsl_linalg_SV_decomp_mod (A,X, V, S, work);
else gsl_linalg_SV_decomp_jacobi(A,V,S);

//unpack the results and write the output file
for (i=0; i<nbasis; i++) {
	EMData out;
	out.setSize(imsize,imsize,1);
	circularunpack(A,&out,i,mask);
	out.writeIMAGIC(argv[2],i);
}

gsl_vector_fprintf(stdout,S,"%g");

}

int circularcount(int r) {
int x,y;

int ret=0;
for (y=-r; y<=r; y++) {
	for (x=-r; x<=r; x++) {
		if (floor(hypot(x,y)+.5)<=r) ret++;
	}
}

return ret;
}

void circularpack(EMData *src,gsl_matrix *dest,int mxy,int r) {
int x,y;

int n=0;
for (y=-r; y<=r; y++) {
	for (x=-r; x<=r; x++) {
		if (floor(hypot(x,y)+.5)>r) continue;
		gsl_matrix_set(dest,n,mxy,src->valueAt(x+src->xSize()/2,y+src->ySize()/2));
		n++;
	}
}


}

void circularunpack(gsl_matrix *src, EMData *dest, int mxy,int r) {
int x,y;

dest->zero();

int n=0;
for (y=-r; y<=r; y++) {
	for (x=-r; x<=r; x++) {
		if (floor(hypot(x,y)+.5)>r) continue;
		dest->setValueAt(x+dest->xSize()/2,y+dest->ySize()/2,0,gsl_matrix_get(src,n,mxy));
		n++;
	}
}
dest->update();
}
